#include "nvfs.h"

static const struct super_operations nvfs_sops;
static struct kmem_cache *nvfs_inode_cache;

static int persistent_memory_claim(struct nvfs_superblock *nvs)
{
	int r;
	loff_t s;
	long p, da;
	pfn_t pfn;
	int id;
	struct page **pages;
	sector_t offset;

	nvs->memory_map_size = i_size_read(nvs->s->s_bdev->bd_inode);
	s = nvs->memory_map_size;
	p = s >> PAGE_SHIFT;
	if (!p) {
		r = -EINVAL;
		goto err1;
	}
	if (p != s >> PAGE_SHIFT) {
		r = -EOVERFLOW;
		goto err1;
	}

	offset = get_start_sect(nvs->s->s_bdev);
	if (offset & (PAGE_SIZE / 512 - 1)) {
		r = -EINVAL;
		goto err1;
	}
	offset >>= PAGE_SHIFT - 9;

	nvs->dax_dev = fs_dax_get_by_bdev(nvs->s->s_bdev);
	if (!nvs->dax_dev) {
		r = -EOPNOTSUPP;
		goto err1;
	}

	id = dax_read_lock();

	if (!dax_synchronous(nvs->dax_dev)) {
		r = -EOPNOTSUPP;
		goto err2;
	}

	da = dax_direct_access(nvs->dax_dev, offset, p, &nvs->memory_map, &pfn);
	if (da < 0) {
		nvs->memory_map = NULL;
		r = da;
		goto err2;
	}
	if (!pfn_t_has_page(pfn)) {
		nvs->memory_map = NULL;
		r = -EOPNOTSUPP;
		goto err2;
	}
	if (da != p) {
		long i;
		nvs->memory_map = NULL;
		pages = kvmalloc_array(p, sizeof(struct page *), GFP_KERNEL);
		if (!pages) {
			r = -ENOMEM;
			goto err2;
		}
		i = 0;
		do {
			long daa;
			daa = dax_direct_access(nvs->dax_dev, offset + i, p - i, NULL, &pfn);
			if (daa <= 0) {
				r = daa ? daa : -EINVAL;
				goto err3;
			}
			if (!pfn_t_has_page(pfn)) {
				r = -EOPNOTSUPP;
				goto err3;
			}
			while (daa-- && i < p) {
				pages[i++] = pfn_t_to_page(pfn);
				pfn.val++;
				if (!(i & 15))
					cond_resched();
			}
		} while (i < p);
		nvs->memory_map = vmap(pages, p, VM_MAP, PAGE_KERNEL);
		if (!nvs->memory_map) {
			r = -ENOMEM;
			goto err3;
		}
		kvfree(pages);
		nvs->memory_vmapped = true;
	}

	dax_read_unlock(id);

	return 0;
err3:
	kvfree(pages);
err2:
	dax_read_unlock(id);
err1:
	return r;
}

__cold static void persistent_memory_release(struct nvfs_superblock *nvs)
{
	if (nvs->memory_vmapped)
		vunmap(nvs->memory_map);
	if (nvs->dax_dev)
		put_dax(nvs->dax_dev);
}

void nvfs_error(struct nvfs_superblock *nvs, const char *fmt, ...)
{
	static DEFINE_RATELIMIT_STATE(nvfs_error_ratelimit, 5 * HZ, 10);
	struct va_format vaf;
	va_list args;

	p_superblock(nvs)->error = cpu_to_le32(1);

	if (!__ratelimit(&nvfs_error_ratelimit))
		goto skip_print;

	va_start(args, fmt);
	vaf.fmt = fmt;
	vaf.va = &args;

	printk(KERN_CRIT "NVFS (%s): error: %pV\n", nvs->s->s_id, &vaf);
	dump_stack();

	va_end(args);

skip_print:
	return;
}

static int nvfs_sync_fs(struct super_block *sb, int wait)
{
	dquot_writeback_dquots(sb, -1);
	return 0;
}

static int nvfs_statfs(struct dentry *dentry, struct kstatfs *buf)
{
	struct nvfs_superblock *nvs = dentry->d_sb->s_fs_info;

	uint64_t fsid;

	buf->f_type = le32_to_cpu(*(uint32_t *)NVFS_MAGIC);
	buf->f_bsize = 1UL << nvs->log2_page_size;
	buf->f_blocks = le64_to_cpu(p_superblock(nvs)->n_pages);
	buf->f_bfree = percpu_counter_sum_positive(&nvs->free_blocks_cnt);
	buf->f_bavail = buf->f_bfree;
	buf->f_files = nvs->n_groups << nvs->log2_inodes_per_group;
	buf->f_ffree = percpu_counter_sum_positive(&nvs->free_inodes_cnt);
	buf->f_namelen = nvfs_max_filename_length(nvs);

	fsid =	le64_to_cpu(*(uint64_t *)(p_superblock(nvs)->uuid)) ^
		le64_to_cpu(*(uint64_t *)(p_superblock(nvs)->uuid + 8));
	buf->f_fsid.val[0] = fsid & 0xFFFFFFFFUL;
	buf->f_fsid.val[1] = (fsid >> 32) & 0xFFFFFFFFUL;

	return 0;
}

#ifdef CONFIG_QUOTA

__cold static int nvfs_quota_on(struct super_block *sb, int type, int format_id, const struct path *path)
{
	struct inode *inode;
	int r;

	r = dquot_quota_on(sb, type, format_id, path);
	if (r)
		return r;

	inode = d_inode(path->dentry);
	inode_lock(inode);
	inode_set_flags(inode, S_NOATIME | S_IMMUTABLE, S_NOATIME | S_IMMUTABLE);
	inode_unlock(inode);

	return 0;
}

__cold static int nvfs_quota_off(struct super_block *sb, int type)
{
	struct inode *inode = sb_dqopt(sb)->files[type];
	int r;

	if (!inode || !igrab(inode))
		goto out;

	r = dquot_quota_off(sb, type);
	if (r)
		goto out_put;

	inode_lock(inode);
	inode_set_flags(inode, 0, S_NOATIME | S_IMMUTABLE);
	inode_unlock(inode);

out_put:
	iput(inode);
	return r;
out:
	return dquot_quota_off(sb, type);
}

static struct dquot **nvfs_get_dquots(struct inode *inode)
{
	struct nvfs_memory_inode *nmi = i_to_nmi(inode);
	return nmi->i_dquot;
}

static const struct quotactl_ops nvfs_quotactl_ops = {
	.quota_on	= nvfs_quota_on,
	.quota_off	= nvfs_quota_off,
	.quota_sync	= dquot_quota_sync,
	.get_state	= dquot_get_state,
	.set_info	= dquot_set_dqinfo,
	.get_dqblk	= dquot_get_dqblk,
	.set_dqblk	= dquot_set_dqblk,
	.get_nextdqblk	= dquot_get_next_dqblk,
};

#endif

static int nvfs_show_options(struct seq_file *seq, struct dentry *root)
{
	struct nvfs_superblock *nvs = root->d_sb->s_fs_info;

	if (nvs->mount_flags & NVFS_USR_QUOTA)
		seq_puts(seq, ",usrquota");
	if (nvs->mount_flags & NVFS_GRP_QUOTA)
		seq_puts(seq, ",grpquota");

	return 0;
}

static void nvfs_inode_ctor(void *n)
{
	struct nvfs_memory_inode *nmi = n;
	init_rwsem(&nmi->dax_sem);
	init_rwsem(&nmi->xattr_sem);
	inode_init_once(&nmi->vfs_inode);
}

static struct inode *nvfs_cache_alloc_inode(struct super_block *sb)
{
	struct nvfs_memory_inode *nmi;
	nmi = kmem_cache_alloc(nvfs_inode_cache, GFP_NOIO);
	if (unlikely(!nmi))
		return NULL;

#ifdef CONFIG_QUOTA
	memset(&nmi->i_dquot, 0, sizeof(nmi->i_dquot));
#endif

	return &nmi->vfs_inode;
}

static void nvfs_cache_free_inode(struct inode *inode)
{
	struct nvfs_memory_inode *nmi = i_to_nmi(inode);
	kmem_cache_free(nvfs_inode_cache, nmi);
}

static void nvfs_flush_cached_data(struct nvfs_superblock *nvs)
{
	uint64_t group;
	for (group = 0; group < nvs->n_groups; group++) {
		__attribute((unused)) struct nvfs_group *p = p_group(nvs, group);
		nv_flush_cached(p, sizeof(struct nvfs_group));
		cond_resched();
	}
	pmem_wmb();
}

enum {
	Opt_usrquota, Opt_grpquota,
};

static match_table_t tokens = {
	{Opt_usrquota, "usrquota"},
	{Opt_grpquota, "grpquota"},
};

__cold static int parse_options(struct nvfs_superblock *nvs, char *options, bool remount)
{
	char *p;
	substring_t args[MAX_OPT_ARGS];

	if (!options)
		return 0;

	while ((p = strsep(&options, ",")) != NULL) {
		int token;
		if (!*p)
			continue;
		token = match_token(p, tokens, args);
		switch (token) {
			case Opt_usrquota:
				if (remount)
					break;
				nvs->mount_flags |= NVFS_USR_QUOTA;
				break;
			case Opt_grpquota:
				if (remount)
					break;
				nvs->mount_flags |= NVFS_GRP_QUOTA;
				break;
			default:
				return -EINVAL;
		}
	}

	return 0;
}

static int nvfs_remount(struct super_block *s, int *flags, char *options)
{
	struct nvfs_superblock *nvs = s->s_fs_info;
	int r;

	r = parse_options(nvs, options, true);
	if (r) {
		return r;
	}
	if ((bool)(*flags & SB_RDONLY) != sb_rdonly(s)) {
		if (sb_rdonly(s)) {
			nvs->saved_dirty = le32_to_cpu(p_superblock(nvs)->dirty);
			nv_store(p_superblock(nvs)->dirty, cpu_to_le32(1));
			pmem_wmb();
			dquot_resume(s, -1);
		} else {
			r = dquot_suspend(s, -1);
			if (r < 0)
				return r;
			nvfs_flush_cached_data(nvs);
			nv_store(p_superblock(nvs)->dirty, nvs->saved_dirty);
		}
	}

	return 0;
}

__cold static void nvfs_free_super(struct nvfs_superblock *nvs)
{
	if (!nvs)
		return;
	persistent_memory_release(nvs);
	percpu_counter_destroy(&nvs->free_blocks_cnt);
	percpu_counter_destroy(&nvs->free_inodes_cnt);
	free_percpu(nvs->npc);
	kvfree(nvs->group_info);
	kfree(nvs);
}

__cold static void nvfs_put_super(struct super_block *s)
{
	struct nvfs_superblock *nvs = s->s_fs_info;

#ifdef CONFIG_QUOTA
	int type;
	for (type = 0; type < MAXQUOTAS; type++)
		nvfs_quota_off(s, type);
#endif

	if (!sb_rdonly(s)) {
		nvfs_flush_cached_data(nvs);
		nv_store(p_superblock(nvs)->dirty, nvs->saved_dirty);
	}

	nvfs_free_super(nvs);
}

__cold static int nvfs_fill_super(struct super_block *s, void *options, int silent)
{
	struct nvfs_superblock *nvs;
	struct inode *root;
	int r;
	unsigned c;
	unsigned long i;
	uint64_t max_bytes;
	const char *msg;
	uint64_t max_ino, max_block;
	uint64_t free_blocks, free_inodes;

	nvs = kzalloc(sizeof(struct nvfs_superblock), GFP_KERNEL);
	if (!nvs) {
		r = -ENOMEM;
		goto err0;
	}

	nvs->s = s;
	s->s_fs_info = nvs;
	s->s_op = &nvfs_sops;
#ifdef CONFIG_NVFS_POSIX_ACL
	s->s_flags |= SB_POSIXACL;
#endif
	s->s_xattr = nvfs_xattr_handlers;
	s->s_time_min = S64_MIN;
	s->s_time_max = S64_MAX;
	s->s_time_gran = 1;
	s->s_max_links = UINT_MAX;

	r = persistent_memory_claim(nvs);
	if (r)
		goto err0;

	msg = nvfs_validate_super(nvs->memory_map, nvs->memory_map_size);
	if (msg) {
		printk(KERN_ERR "NVFS (%s): %s\n", nvs->s->s_id, msg);
		r = -EINVAL;
		goto err0;
	}

	nvfs_get_sizes(p_superblock(nvs), &nvs->n_groups, NULL, NULL, NULL, NULL, &max_bytes);

	nvs->log2_page_size = p_superblock(nvs)->log2_page_size;
	nvs->log2_inode_size = p_superblock(nvs)->log2_inode_size;
	nvs->log2_bytes_per_group = p_superblock(nvs)->log2_page_size * 2 + 3;
	nvs->log2_inodes_per_group = p_superblock(nvs)->log2_inode_pages_per_group + p_superblock(nvs)->log2_page_size - p_superblock(nvs)->log2_inode_size;

	s->s_blocksize_bits = nvs->log2_page_size;
	s->s_maxbytes = min(max_bytes, (uint64_t)MAX_LFS_FILESIZE);

	//printk("s->s_maxbytes: %llx %lld\n", s->s_maxbytes, s->s_maxbytes);

	nvs->npc = alloc_percpu(struct nvfs_per_cpu);
	if (!nvs->npc) {
		r = -ENOMEM;
		goto err0;
	}
	max_ino = nvs->n_groups << nvs->log2_inodes_per_group;
	max_block = nvs->n_groups << log2_pages_per_group(nvs);
	for_each_possible_cpu(c) {
		struct nvfs_per_cpu *npc = per_cpu_ptr(nvs->npc, c);
		npc->current_inode = nvfs_get_random(max_ino);
		npc->current_block = nvfs_get_random(max_block);
	}

	nvs->group_info = kvmalloc_array(nvs->n_groups, sizeof(struct nvfs_group_info), GFP_KERNEL);
	if (!nvs->group_info) {
		r = -ENOMEM;
		goto err0;
	}

	free_blocks = free_inodes = 0;
	for (i = 0; i < nvs->n_groups; i++) {
		struct nvfs_group_info *gi = nvs->group_info + i;
		struct nvfs_group *g = p_group(nvs, i);

		gi->n_free_blocks = le32_to_cpu(g->n_free_blocks);
		gi->n_free_inodes = le32_to_cpu(g->n_free_inodes);
		free_blocks += gi->n_free_blocks;
		free_inodes += gi->n_free_inodes;

		spin_lock_init(&gi->lock);

		cond_resched();
	}

	r = percpu_counter_init(&nvs->free_blocks_cnt, free_blocks, GFP_KERNEL);
	if (r) {
		r = -ENOMEM;
		goto err0;
	}
	r = percpu_counter_init(&nvs->free_inodes_cnt, free_inodes, GFP_KERNEL);
	if (r) {
		r = -ENOMEM;
		goto err0;
	}

	r = parse_options(nvs, options, false);
	if (r) {
		goto err0;
	}

#ifdef CONFIG_QUOTA
	s->dq_op = &dquot_operations;
	s->s_qcop = &nvfs_quotactl_ops;
	s->s_quota_types = QTYPE_MASK_USR | QTYPE_MASK_GRP;
#endif

	root = nvfs_iget(s, le64_to_cpu(p_superblock(nvs)->root_inode));
	if (IS_ERR(root)) {
		r = PTR_ERR(root);
		goto err0;
	}
	if (!S_ISDIR(root->i_mode)) {
		iput(root);
		r = -EUCLEAN;
		goto err0;
	}
	s->s_root = d_make_root(root);
	if (!s->s_root) {
		r = -ENOMEM;
		goto err0;
	}

	nvs->saved_dirty = le32_to_cpu(p_superblock(nvs)->dirty);
	if (!sb_rdonly(s))
		nv_store(p_superblock(nvs)->dirty, cpu_to_le32(1));
	pmem_wmb();

	return 0;

err0:
	nvfs_free_super(nvs);
	return r;
}

__cold static struct dentry *nvfs_mount(struct file_system_type *fs_type, int flags,
					const char *dev_name, void *data)
{
	return mount_bdev(fs_type, flags, dev_name, data, nvfs_fill_super);
}

static const struct super_operations nvfs_sops = {
	.alloc_inode = nvfs_cache_alloc_inode,
	.free_inode = nvfs_cache_free_inode,
	.evict_inode = nvfs_evict_inode,
	.put_super = nvfs_put_super,
	.sync_fs = nvfs_sync_fs,
	.statfs  = nvfs_statfs,
	.show_options = nvfs_show_options,
	.remount_fs = nvfs_remount,
#ifdef CONFIG_QUOTA
	.quota_read = nvfs_quota_read,
	.quota_write = nvfs_quota_write,
	.get_dquots = nvfs_get_dquots,
#endif
};

static struct file_system_type nvfs_fs_type = {
	.owner = THIS_MODULE,
	.name = "nvfs",
	.mount = nvfs_mount,
	.kill_sb = kill_block_super,
	.fs_flags = FS_REQUIRES_DEV
};

__cold static int __init nvfs_init(void)
{
	int r;

	nvfs_inode_cache = kmem_cache_create("nvfs_inode_cache", sizeof(struct nvfs_memory_inode), 0, SLAB_RECLAIM_ACCOUNT | SLAB_MEM_SPREAD | SLAB_ACCOUNT, nvfs_inode_ctor);
	if (!nvfs_inode_cache) {
		r = -ENOMEM;
		goto ret_r;
	}

	r = register_filesystem(&nvfs_fs_type);
	if (unlikely(r))
		goto ret_r;

	return 0;

ret_r:
	kmem_cache_destroy(nvfs_inode_cache);
	return r;
}

__cold static void __exit nvfs_exit(void)
{
	unregister_filesystem(&nvfs_fs_type);
	rcu_barrier();
	kmem_cache_destroy(nvfs_inode_cache);
}

module_init(nvfs_init)
module_exit(nvfs_exit)

MODULE_ALIAS_FS("nvfs");

MODULE_AUTHOR("Mikulas Patocka");
MODULE_DESCRIPTION("NVFS Filesystem");
MODULE_LICENSE("GPL");
